/*
  Copyright Red Hat, Inc. 2004

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2, or (at your option) any
  later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to the
  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
  MA 02139, USA.
*/
//#define DEBUG
#include <magma.h>
#include <magmamsg.h>
#include <resgroup.h>
#include <rg_queue.h>
#include <platform.h>
#include <msgsimple.h>
#include <clulog.h>


struct fw_message {
	SmMessageSt msg;
	uint64_t nodeid;
	int fd;
	int unused;
};


void
build_message(SmMessageSt *msgp, int action, char *svcName, uint64_t target)
{
	msgp->sm_hdr.gh_magic = GENERIC_HDR_MAGIC;
	msgp->sm_hdr.gh_command = RG_ACTION_REQUEST;
	msgp->sm_hdr.gh_length = sizeof(*msgp);
	msgp->sm_data.d_action = action;
	strncpy(msgp->sm_data.d_svcName, svcName,
		sizeof(msgp->sm_data.d_svcName));
	msgp->sm_data.d_svcOwner = target;
	msgp->sm_data.d_ret = 0;

	swab_SmMessageSt(msgp);
}


void *
forwarding_thread(void *arg)
{
	rg_state_t rgs;
	request_t *req = (request_t *)arg;
	void *lockp;
	int fd, ret;
	SmMessageSt msg;
	cluster_member_list_t *m = NULL;

	if (rg_lock(req->rr_group, &lockp) != 0) {
		msg_close(req->rr_resp_fd);
		rq_free(req);
		pthread_exit(NULL);
	}

	if (get_rg_state(req->rr_group, &rgs) != 0) {
		rg_unlock(req->rr_group, lockp);
		msg_close(req->rr_resp_fd);
		rq_free(req);
		pthread_exit(NULL);
	}

	rg_unlock(req->rr_group, lockp);

	/* Construct message */
	build_message(&msg, req->rr_request, req->rr_group, req->rr_target);

	/* 
	clulog(LOG_DEBUG, "Forwarding %s request to %d\n",
	       rg_req_str(req->rr_request), (int)rgs.rs_owner);
	 */

	fd = msg_open(rgs.rs_owner, RG_PORT, RG_PURPOSE, 10);
	if (fd == -1) {
		msg_close(req->rr_resp_fd);
		rq_free(req);
		pthread_exit(NULL);
	}

	if (msg_send(fd, &msg, sizeof(msg)) != sizeof(msg)) {
		msg_close(fd);
		msg_close(req->rr_resp_fd);
		rq_free(req);
		pthread_exit(NULL);
	}

	/*
	 * Ok, we're forwarding a message to another node.  Keep tabs on 
	 * the node to make sure it doesn't die.  Basically, wake up every
	 * now and again to make sure it's still online.  If it isn't, send
	 * a response back to the caller.
	 */
	do {
		ret = msg_receive_timeout(fd, &msg, sizeof(msg), 10);
		if (ret < (int)sizeof(msg)) {
			if (ret < 0 && errno == ETIMEDOUT) {
				m = member_list();
				if (!memb_online(m, rgs.rs_owner)) {
					msg.sm_data.d_ret = RG_ENODEDEATH;
					/* we decode down below,
					 * so encode here */
					swab_SmMessageSt(&msg);
					break;
				}
				cml_free(m);
				m = NULL;
				continue;
			}
			msg_close(fd);
			msg_close(req->rr_resp_fd);
			goto out;
		}
		break;
	} while(1);

	if (m)
		cml_free(m);
	msg_close(fd);

	swab_SmMessageSt(&msg);
	send_response(msg.sm_data.d_ret, req->rr_target, req);
out:
	rq_free(req);
	pthread_exit(NULL);
}


void
forward_request(request_t *req)
{
	pthread_t newthread;
	pthread_attr_t attrs;

        pthread_attr_init(&attrs);
        pthread_attr_setinheritsched(&attrs, PTHREAD_INHERIT_SCHED);
        pthread_attr_setdetachstate(&attrs, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attrs, 262144);

	pthread_create(&newthread, &attrs, forwarding_thread, req);
        pthread_attr_destroy(&attrs);
}


void *
forwarding_thread_v2(void *arg)
{
	int fd = -1, resp_fd = -1;
	cluster_member_list_t *m = NULL;
	SmMessageSt *msgp = NULL, msg;
	int response_code = RG_EAGAIN, ret, target = -1;
	int retries = 0;
	struct fw_message *fwmsg = (struct fw_message *)arg;

	msgp = &fwmsg->msg;
	resp_fd = fwmsg->fd;
	target = fwmsg->nodeid;

	clulog(LOG_DEBUG, "FW: Forwarding SM request to %d\n",
	       target);

	if ((fd = msg_open(target, RG_PORT, RG_PURPOSE, 10)) < 0) {
		clulog(LOG_DEBUG, "FW: Failed to open channel to %d: %s\n",
		       target, strerror(errno));
		goto out_fail;
	}

	/* swap + send */
	swab_SmMessageSt(msgp);
	if (msg_send(fd, msgp, sizeof(*msgp)) < sizeof(*msgp)) {
		clulog(LOG_DEBUG, "FW: Failed to send message to %d fd %d: %s\n",
		       target, fd, strerror(errno));
		goto out_fail;
	}


        /*
	 * Ok, we're forwarding a message to another node.  Keep tabs on
	 * the node to make sure it doesn't die.  Basically, wake up every
	 * now and again to make sure it's still online.  If it isn't, send
	 * a response back to the caller.
	 */
	do {
		ret = msg_receive_timeout(fd, &msg, sizeof(msg), 10);
		if (ret < (int)sizeof(msg)) {
			if (ret < 0 && errno == ETIMEDOUT) {
				if (!member_online(target)) {
					response_code = RG_ENODE;
					goto out_fail;
				}
				continue;
			}

			if (ret == 0)
				continue;
		}
		break;
	} while(++retries < 60); /* old 600 second rule */

	swab_SmMessageSt(&msg);

	response_code = msg.sm_data.d_ret;
	target = msg.sm_data.d_svcOwner;

out_fail:
	free(fwmsg); 

	if (resp_fd >= 0) {
		send_ret(resp_fd, msgp->sm_data.d_svcName, response_code,
			 msgp->sm_data.d_action, target);
		msg_close(resp_fd);
	}

	if (fd >= 0)
		msg_close(fd);

	pthread_exit(NULL);
}


void
forward_message(int fd, void *msgp, uint64_t nodeid)
{
	pthread_t newthread;
	pthread_attr_t attrs;
	struct fw_message *fwmsg;

	fwmsg = malloc(sizeof(struct fw_message));
	if (!fwmsg) {
		msg_close(fd);
		return;
	}

	memcpy(&fwmsg->msg, msgp, sizeof(fwmsg->msg));
	fwmsg->fd = fd;
	fwmsg->nodeid = nodeid;

        pthread_attr_init(&attrs);
        pthread_attr_setinheritsched(&attrs, PTHREAD_INHERIT_SCHED);
        pthread_attr_setdetachstate(&attrs, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attrs, 262144);

	pthread_create(&newthread, &attrs, forwarding_thread_v2, fwmsg);
        pthread_attr_destroy(&attrs);
}
